import logging

from flask_jwt_extended import jwt_required
from flask_restplus import Namespace, Resource

from portal import storage
from portal.rest import rest_helpers
from portal.rest.schemas import review_schema, submission_schema, submission_state_schema
from portal.service import auth, general, permissions
from portal.service.permissions import check_client, require_client
from portal.service.reviews import create_review, create_review_items
from portal.service.submissions import copy_submission, delete_submission, \
    update_submission_state, cancel_submission, send_file_or_zip, upload_results_to_storage

submissions_namespace = Namespace('submissions')

log = logging.getLogger(__name__)


def submission_access_group(client, submission, course, perm):
    if permissions.check_client(client, course, perm):
        group_intersection = [
            group for group in client.groups if submission.user in group.users]
        return any(
            group in submission.project.groups for group in group_intersection)
    return False


@submissions_namespace.route('/<string:sid>')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResource(Resource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submission found', model=submission_schema)
    def get(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        course = submission.project.course
        # authorization
        checks = [
            check_client(client, course, ['read_submissions_all']),
            submission_access_group(client, submission, course, [
                'read_submissions_groups']),
            (check_client(client, course, [
                'read_submissions_own']) and submission.user == client)
            ]
        permissions.require_any_check(client, checks=checks)

        return submission_schema.dump(submission)

    @jwt_required
    @submissions_namespace.response(204, 'Submission deleted')
    def delete(self, sid: str):
        client = auth.find_client()
        permissions.require_sysadmin(client)
        submission = general.find_submission(sid)
        delete_submission(submission)
        return '', 204


@submissions_namespace.route('/<string:sid>/state')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionState(Resource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submission state', model=submission_state_schema)
    def get(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        course = submission.project.course
        # authorization
        checks = [
            permissions.check_client(client, course, ['read_submissions_all']),
            submission_access_group(client, submission, course, [
                'read_submissions_groups']),
            (check_client(client, course, [
                'read_submissions_own']) and submission.user == client)
            ]

        permissions.require_any_check(client, checks=checks)

        return submission_state_schema.dump(submission)[0]

    @jwt_required
    @submissions_namespace.response(204, 'Submission state updated')
    def put(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        # authorization
        checks = [
            permissions.check_component(component=client),
            permissions.check_sysadmin(client),
            client == submission.user,
            ]
        permissions.require_any_check(client, checks=checks)

        data = rest_helpers.parse_request_data(
            submission_state_schema, action='update_state', resource='submission'
            )
        update_submission_state(client, submission, data)
        return '', 204


@submissions_namespace.route('/<string:sid>/result')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResult(Resource):
    @jwt_required
    def put(self, sid: str):
        raise NotImplementedError()


@submissions_namespace.route('/<string:sid>/files')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionFiles(Resource):
    @jwt_required
    def get(self, sid: str):
        raise NotImplementedError()


@submissions_namespace.route('/<string:sid>/files/sources')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionSourceFiles(Resource):
    @jwt_required
    def get(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        course = submission.project.course
        # authorization
        checks = [
            check_client(client, course, ['read_submissions_all']),
            submission_access_group(client, submission, course, [
                'read_submissions_groups']),
            (check_client(client, course, [
                'read_submissions_own']) and submission.user == client)
            ]
        permissions.require_any_check(client, checks=checks)
        storage_submission = storage.submissions.get(submission.id)
        return send_file_or_zip(storage_submission)


@submissions_namespace.route('/<string:sid>/files/test_files')
@submissions_namespace.doc({'sid': 'Submission id'})
class SubmissionTestFiles(Resource):
    @jwt_required
    def get(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        course = submission.project.course
        # authorization
        checks = [
            check_client(client, course, ['read_submissions_all']),
            submission_access_group(client, submission, course, [
                'read_submissions_groups'])
            ]
        permissions.require_any_check(client, checks=checks)
        storage_entity = storage.test_files.get(submission.project.id)
        return send_file_or_zip(storage_entity)


@submissions_namespace.route('/<string:sid>/files/results')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResultFiles(Resource):
    @jwt_required
    def get(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        course = submission.project.course
        # authorization
        checks = [
            check_client(client, course, ['read_submissions_all']),
            submission_access_group(client, submission, course, [
                'read_submissions_groups'])
            ]
        permissions.require_any_check(client, checks=checks)
        storage_entity = storage.results.get(submission.id)
        return send_file_or_zip(storage_entity)

    @jwt_required
    def post(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        # authorization
        permissions.require_sysadmin(client)
        permissions.require_any_check(client, checks=[permissions.check_sysadmin(client),
                                                      permissions.check_component(client)])

        task = upload_results_to_storage(submission)
        return {'new_task': task}


@submissions_namespace.route('/<string:sid>/resubmit')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResubmit(Resource):
    @jwt_required
    # @submissions_namespace.response(201, 'New submission', model=submission_schema)
    def post(self, sid: str):
        client = auth.find_client()
        source_submission = general.find_submission(sid)
        # authorization
        course = source_submission.project.course
        require_client(client=client, course=course,
                       permissions=['resubmit_submissions'])

        data = rest_helpers.parse_request_data(
            submission_schema, action='resubmit', resource='submission'
            )
        # create new submission by copying files from the source submission in
        # storage
        new_submission = copy_submission(source_submission, note=data['note'])
        return submission_schema.dump(new_submission), 201


@submissions_namespace.route('/<string:sid>/cancel')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
@submissions_namespace.response(204, 'Submissions canceled')
class SubmissionCancel(Resource):
    @jwt_required
    # @submissions_namespace.response(201, 'New submission', model=submission_schema)
    def post(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        # authorization
        course = submission.project.course
        require_client(client=client, course=course,
                       permissions=['resubmit_submissions'])
        # create new submission by copying files from the source submission in
        # storage
        cancel_submission(submission)
        return '', 204


@submissions_namespace.route('/<string:sid>/review')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionReview(Resource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submissions review', model=review_schema)
    def get(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        course = submission.project.course
        # authorization
        checks = [
            permissions.check_client(client, course, ['read_submissions_all']),
            submission_access_group(client, submission, course, [
                'read_submissions_groups']),
            (check_client(client, course, [
                'read_submissions_own']) and submission.user == client)
            ]
        permissions.require_any_check(client, checks=checks)

        return review_schema.dump(submission.review)[0]

    @jwt_required
    # @submissions_namespace.response(201, 'Submissions review created', model=review_schema)
    def post(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        course = submission.project.course
        # authorization
        checks = [
            permissions.check_client(client, course, ['write_reviews_all']),
            submission_access_group(client, submission, course, [
                'write_reviews_group']),
            (check_client(client, course, [
                'write_reviews_own']) and submission.user == client)
            ]
        permissions.require_any_check(client, checks=checks)

        data = rest_helpers.parse_request_data(
            review_schema, action='create', resource='review')

        if not submission.review:
            create_review(submission)

        # also writes to db
        create_review_items(review=submission.review,
                            items=data['review_items'], author=client)

        return review_schema.dump(submission.review)[0], 201
